﻿using System;
using System.Reflection;
using System.Text;
using DynamicProxy;

namespace WdtClient
{
    /// <summary>
    /// 声明API Attribute
    /// </summary>
    [AttributeUsage(AttributeTargets.Method)]
    public class ApiAttribute : Attribute
    {
        public ApiAttribute(string api, bool paged = false)
        {
            this.api = api;
            this.paged = paged;
        }

        public string GetApi()
        {
            return this.api;
        }

        public bool GetPaged()
        {
            return this.paged;
        }

        // api名
        private string api;
        // 是否分页
        private bool paged;
    }

    /// <summary>
    /// 分页查询参数
    /// </summary>
    public class Pager
    {
	    private int pageSize = 0;
	    private int pageNo = 0;
	    private bool calcTotal = false;
	
	    public Pager(int pageSize, int pageNo = 0, bool calcTotal = false)
	    {
		    this.pageSize = pageSize;
		    this.pageNo = pageNo;
		    this.calcTotal = calcTotal;
	    }
	    
        /// <summary>
        /// 分页,每页条数
        /// </summary>
        public int PageSize
        {
            get { return this.pageSize; }
            set { this.pageSize = value; }
        }
        
        /// <summary>
        /// 页码,从0开始
        /// </summary>
        public int PageNo
        {
            get { return this.pageNo; }
            set { this.pageNo = value; }
        }

        /// <summary>
        /// 是否返回最条数, true时效率比较低, 尽量避免
        /// </summary>
        public bool CalcTotal
        {
            get { return this.calcTotal; }
            set { this.calcTotal = value; }
        }
    }

    /// <summary>
    /// 接口异常返回
    /// </summary>
    public class WdtErpException : Exception
    {
        private int code;

        public WdtErpException(int code, string message) : base(message)
        {
            this.code = code;
        }

        public int Code
        {
            get { return this.code; }
        }
    }

    public interface IClient
    {
        /// <summary>
        /// 网络超时时间, 单位毫秒
        /// </summary>
        int Timeout
        {
            get;
            set;
        }

        /// <summary>
        /// 执行调用
        /// </summary>
        /// <param name="method">方法名,如setting.Shop.get</param>
        /// <param name="args">调用参数</param>
        /// <param name="pager">分页参数</param>
        /// <param name="returnType">返回值类型</param>
        /// <returns>返回值</returns>
        object Excecute(string method, object[] args, Pager pager, Type returnType);
    }

    public static class ApiFactory<T> where T : class
    {
        private class Proxy : IProxyInvocationHandler
        {
            private IClient client;

            public Proxy(IClient client)
            {
                this.client = client;
            }

            public Object Invoke(Object proxy, MethodInfo method, Object[] parameters)
            {
                ApiAttribute api = (ApiAttribute) System.Attribute.GetCustomAttribute(method, typeof(ApiAttribute));
                if (api == null)
                    throw new WdtErpException(-1, "Api Name not indicated");

                Pager pager = null;
                // 如果分页查询, 最后一个参数是pager
                if (api.GetPaged())
                {
                    int len = parameters.Length;
                    pager = (Pager)parameters[len - 1];

                    object[] newParams = new object[len - 1];
                    Array.Copy(parameters, 0, newParams, 0, len - 1);
                    parameters = newParams;
                }

                return this.client.Excecute(api.GetApi(), parameters, pager, method.ReturnType);
            }
        }

        public static T Get(IClient client)
        {
            return (T) ProxyFactory.GetInstance().Create(new Proxy(client), typeof(T), true);
        }
    }

    public static class ExApiFactory<T> where T : class
    {
        private class Proxy : IProxyInvocationHandler
        {
            private IClient client;

            public Proxy(IClient client)
            {
                this.client = client;
            }

            public Object Invoke(Object proxy, MethodInfo method, Object[] parameters)
            {
                ApiAttribute api = (ApiAttribute)System.Attribute.GetCustomAttribute(method, typeof(ApiAttribute));
                if (api == null)
                    throw new WdtErpException(-1, "Api Name not indicated");

                if(api.GetPaged())
                    throw new WdtErpException(-2, "Pagination is not supported for Script Api");

                int len = parameters.Length;

                object[] newParams = new object[len + 1];
                newParams[0] = api.GetApi();
                Array.Copy(parameters, 0, newParams, 1, len);

                return this.client.Excecute("system.ScriptExtension.call", newParams, null, method.ReturnType);
            }
        }

        public static T Get(IClient client)
        {
            return (T)ProxyFactory.GetInstance().Create(new Proxy(client), typeof(T), true);
        }
    }
}
