﻿using System;
using System.Collections.Generic;
using System.Collections.Specialized;
using System.Net;
using System.Security.Cryptography;
using System.Text;
using System.IO;
using Newtonsoft.Json;
using Newtonsoft.Json.Linq;
using Newtonsoft.Json.Serialization;

namespace WdtClient.impl
{
    class WebClientEx : WebClient
    {
        private int timeout;

        public int Timeout
        {
            get { return timeout; }
            set { timeout = value; }
        }

        public WebClientEx()
        {
            timeout = -1;
        }

        protected override WebRequest GetWebRequest(Uri address)
        {
            WebRequest request = base.GetWebRequest(address);

            if (request is HttpWebRequest)
            {
                HttpWebRequest httpRequest = request as HttpWebRequest;

                if (timeout > 0)
                    httpRequest.Timeout = timeout;
            }

            return request;
        }
    }

    public class DefaultClient : IClient
    {
        // 协议版本号
        public const string version = "1.0";
        // 时间戳偏移
        private static long TICKS_DIFF = ((new DateTime(1970, 1, 1).Ticks) / 10000000L + 1325347200L);

        private string sid; // 卖家标识
        private string url; // 服务器URL
        private string key;  // openapi key
        private string secret; // openapi secret
        private string salt; // openapi salt
        private int timeout = 30000; // http超时时间, 默认30秒

        // json 序列化选项
        private static JsonSerializerSettings jsonSerializerSettings = new JsonSerializerSettings
        {
            ContractResolver = new DefaultContractResolver
            {
                NamingStrategy = new SnakeCaseNamingStrategy()
            },

            NullValueHandling = NullValueHandling.Ignore
        };

        public DefaultClient(string sid, string url, string key, string secret)
        {
            if (url.EndsWith("/"))
                url += "openapi";
            else
                url += "/openapi";
            this.url = url;

            this.sid = sid;
            this.key = key;

            int pos = secret.IndexOf(':');
            this.secret = secret.Substring(0, pos);
            this.salt = secret.Substring(pos + 1);
        }

        private string ToQueryString(IEnumerable<KeyValuePair<string, string>> args)
        {
            StringBuilder sb = new StringBuilder();

            bool first = true;
            foreach (var item in args)
            {
                if (!first)
                    sb.Append("&");
                else
                    first = false;

                sb.AppendFormat("{0}={1}", Uri.EscapeDataString(item.Key), Uri.EscapeDataString(item.Value));
            }

            return sb.ToString();
        }

        /// <summary>
        /// 计算参数签名
        /// </summary>
        /// <param name="args"></param>
        /// <returns></returns>
        private string Sign(IEnumerable<KeyValuePair<string, string>> args, string secret)
        {
            StringBuilder sb = new StringBuilder();
            sb.Append(secret);
            foreach (var item in args)
            {
                if (item.Key == "sign")
                    continue;

                sb.Append(item.Key);
                sb.Append(item.Value);
            }
            sb.Append(secret);

            MD5 md5 = MD5.Create();

            byte[] md5Hash = md5.ComputeHash(Encoding.UTF8.GetBytes(sb.ToString()));

            sb.Clear();

            foreach (byte b in md5Hash)
                sb.AppendFormat("{0:x2}", b);

            return sb.ToString();
        }

        /// <summary>
        /// 超时时间
        /// </summary>
        public int Timeout
        {
            get { return timeout; }
            set { timeout = value; }
        }

        public object Excecute(string method, object[] args, Pager pager, Type returnType)
        {
            var requestParams = new SortedList<string, string>();
            requestParams.Add("sid", this.sid);
            requestParams.Add("key", this.key);
            requestParams.Add("method", method);
            requestParams.Add("salt", this.salt);
            requestParams.Add("timestamp", (DateTime.UtcNow.Ticks / 10000000L - TICKS_DIFF).ToString());
            requestParams.Add("v", version);

            if (pager != null)
            {
                requestParams.Add("page_size", pager.PageSize.ToString());
                requestParams.Add("page_no", pager.PageNo.ToString());
                requestParams.Add("calc_total", pager.CalcTotal ? "1" : "0");
            }

            // 请求参数
            string body = JsonConvert.SerializeObject(args, jsonSerializerSettings);
            requestParams.Add("body", body);

            // 计算签名
            string sign = this.Sign(requestParams, this.secret);
            requestParams.Remove("body");
            requestParams.Add("sign", sign);

            var requestUrl = this.url + "?" + this.ToQueryString(requestParams);

            string responseBody;
            using (var client = new WebClientEx())
            {
                client.Encoding = Encoding.UTF8;
                client.Timeout = this.timeout;
                client.Headers.Add(HttpRequestHeader.ContentType, "application/json");

                responseBody = client.UploadString(requestUrl, "POST", body);
            }

            var jsonResponse = JObject.Parse(responseBody);
            var status = (int)jsonResponse["status"];
            if (status > 0)
            {
                var message = (string)jsonResponse["message"];
                throw new WdtErpException(status, message);
            }

            if (returnType == typeof(void))
                return null;

            object result = JsonConvert.DeserializeObject(jsonResponse["data"].ToString(), returnType, jsonSerializerSettings);
            if (pager == null || !pager.CalcTotal || result == null)
                return result;

            if (!jsonResponse.ContainsKey("total"))
                return result;

            var totalField = returnType.GetField("Total");
            if (totalField != null)
                totalField.SetValue(result, (int) jsonResponse["total"]);

            return result;
        }

        public static IClient get(string sid, string url, string key, string secret)
        {
            return new DefaultClient(sid, url, key, secret);
        }

        // 服务器地址缓存
        private static string hostUrlCache;
        private static long hostUrlCacheTime;

        public static IClient get(string sid, string key, string secret)
        {
            var now = DateTime.Now.Ticks / 10000000;
            if (!string.IsNullOrEmpty(hostUrlCache))
            {
                if (now - hostUrlCacheTime < 36000)
                    return new DefaultClient(sid, hostUrlCache, key, secret);

                hostUrlCacheTime = 0;
                hostUrlCache = string.Empty;
            }

            string cacheFile = Path.Combine(Path.GetTempPath(), "wdt_host_ip_cache");
            if(File.Exists(cacheFile))
            {
                string cacheData = File.ReadAllText(cacheFile);
                if(!string.IsNullOrEmpty(cacheData))
                {
                    int pos = cacheData.IndexOf(' ');
                    if(pos > 0)
                    {
                        var cacheTime = long.Parse(cacheData.Substring(0, pos));
                        if (now - cacheTime < 36000)
                        {
                            hostUrlCache = cacheData.Substring(pos + 1);
                            hostUrlCacheTime = cacheTime;
                        }
                    }
                }
            }

            if (string.IsNullOrEmpty(hostUrlCache))
            {
                // 访问API
                string response;
                using (var client = new WebClient())
                {
                    client.Encoding = Encoding.UTF8;

                    response = client.DownloadString("http://erp.wangdian.cn/mobile/map.php?sid=" + sid);
                }

                var jsonResponse = JObject.Parse(response);
                var code = (int)jsonResponse["code"];
                if (code != 0)
                {
                    var message = (string)jsonResponse["message"];
                    throw new WdtErpException(code, message);
                }

                var host = (string)jsonResponse["host"];
                hostUrlCache = "http://" + host + "/";
                hostUrlCacheTime = now;

                // 保存缓存
                var cacheData = "" + now + " " + hostUrlCache;
                File.WriteAllText(cacheFile, cacheData);
            }

            return new DefaultClient(sid, hostUrlCache, key, secret);
        }
    }
}
